
#include <msp430.h>
#include <stdint.h>
#include <stdlib.h>
#include "nokia5110.h"
#include "swiic.h"
#include "ina219.h"

												// PORT1
static const unsigned LED			= BIT0;
static const unsigned RXD 			= BIT1;
static const unsigned TXD 			= BIT2;
static const unsigned SWITCH 		= BIT3;
static const unsigned LCD_CLK 		= BIT5;
static const unsigned LCD_BACKLIGHT = BIT6;
static const unsigned LCD_DATA 		= BIT7;
												// PORT2
static const unsigned LCD_DC 		= BIT0;
static const unsigned LCD_CE 		= BIT1;
static const unsigned SCL			= BIT2;
static const unsigned SDA			= BIT3;

												// LCD template
static nokia5110::Nokia5110<P1OUT, LCD_CLK, LCD_DATA, P2OUT, LCD_DC, P2OUT, LCD_CE, 10000> lcd;

typedef CSWIIC<SCL, SDA, P2IN, P2OUT, P2DIR> TIIC;
static TIIC iic;								// IIC template
static CINA219<TIIC, 0x40> ina(iic);			// INA219 template


static const unsigned char ti[] = {				// TI logo
	48, 48 / 8,
	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
	0x00,0x00,0x00,0xFC,0xFC,0xFC,0xFC,0xFC,0xFC,0xFC,
	0xFC,0xFC,0xFC,0xFC,0x00,0x00,0x00,0x00,0x00,0xF0,
	0xF0,0xF0,0xF0,0xF0,0x00,0x00,0x00,0x00,0x00,0x00,
	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
	
	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
	0x00,0x00,0x00,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
	0xFF,0xFF,0xFF,0xFF,0xFC,0xFC,0x3C,0x80,0xFC,0xFD,
	0xFD,0xFD,0x3D,0x81,0xFC,0xFC,0xFC,0xFC,0xFC,0xFC,
	0xFC,0xFC,0xFC,0xFC,0x00,0x00,0x00,0x00,
	
	0x00,0x1C,0x7C,0xFC,0xFC,0xFC,0xFC,0xFC,0xFC,0xFC,
	0xFC,0xFC,0xFC,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
	0xFF,0xFF,0xF7,0xF0,0xF0,0x00,0xF0,0xFF,0xFF,0xFF,
	0xFF,0x0F,0xF0,0xF0,0xF0,0xF0,0xFF,0xFF,0xFF,0xFF,
	0xFF,0xFF,0xFF,0xFF,0xF8,0xF0,0x00,0x00,
	
	0x00,0x00,0x00,0x00,0x01,0x03,0x07,0x3F,0x7F,0xFF,
	0xFF,0xFF,0xFF,0xFF,0x7F,0x3F,0x3F,0x7F,0xFF,0xFF,
	0xFF,0xFF,0xFF,0x1F,0x00,0x1E,0x1F,0x1F,0x1F,0x1F,
	0x01,0x1E,0x1F,0xDF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
	0xFF,0x7F,0x7F,0x3F,0x3F,0x3F,0x00,0x00,
	
	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
	0x01,0x01,0x01,0x00,0x00,0x00,0x00,0x00,0x00,0x03,
	0x0F,0x3F,0xFF,0xFE,0xFE,0xFE,0xFC,0xFC,0xFC,0xFC,
	0xFC,0xFC,0xFC,0x7F,0x1F,0x07,0x03,0x03,0x01,0x00,
	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
	
	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
	0x00,0x00,0x00,0x01,0x07,0x0F,0x0F,0x1F,0x1F,0x3F,
	0x3F,0x3F,0x3F,0x1C,0x00,0x00,0x00,0x00,0x00,0x00,
	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00
};

static const unsigned char AVW[] = {			// V  A  W vertically
	15, 48 / 8,
	0x04,0x1C,0x7C,0xE4,0x80,0x00,0x80,0x64,0x1C,0x04,0x00,0x00,0x00,0x00,0x00, // V
	0x00,0x00,0x00,0x03,0x0F,0x0C,0x03,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
	0x00,0x00,0x80,0x70,0x0C,0x7C,0xF0,0x80,0x00,0x00,0x00,0x00,0x00,0x00,0x00, // A
	0x08,0x0E,0x09,0x01,0x01,0x01,0x09,0x0F,0x0E,0x08,0x00,0x00,0x00,0x00,0x00,
	0x04,0x1C,0xFC,0xE4,0x00,0x80,0x60,0x1C,0xFC,0xE0,0x00,0x80,0x64,0x1C,0x04, // W
	0x00,0x00,0x01,0x0F,0x0E,0x03,0x00,0x00,0x01,0x0F,0x0E,0x03,0x00,0x00,0x00
};

void print_int(int i, unsigned char y)
{												// Print +2.3 integer in large text
	bool neg;									// Make positive if necessary
	if(neg = i < 0) i = -i;						//  and remember if it was negative
	unsigned char x = 10 * (6 - 1) + 6;			// 5 digits, sign, and decimal
	div_t d = div(i, 10); lcd.pd10(d.rem, x, y); x -= 10; // 5th digit
	d = div(d.quot, 10);  lcd.pd10(d.rem, x, y); x -= 16; // 4th digit
	lcd.pd10(11, x, y); x += 6;					// Decimal - this will print 9 wide, but there
												//  is only space for 6, so do before following
												//  digit 
	d = div(d.quot, 10);  lcd.pd10(d.rem, x, y); x -= 16; // 3rd digit
												// 2nd digit if two more digits need to be printed
	if(d.quot > 10) { d = div(d.quot, 10);  lcd.pd10(d.rem, x, y); x -= 10; }
	lcd.pd10(d.quot, x, y); x -= 10;			// 1st or 2nd digit
	lcd.pd10(neg ? 12 : 10, x, y);				// Sign
	if(x >= 10) lcd.pd10(10, x - 10, y);		// Space if necessary
}

void print_reg(unsigned r)
{												// Print register in hex and decimal
	static const char hex[] = "0123456789ABCDEF";
	
	int n = ina.read(r);						// Read the register
	bool neg;									// Make positive if necessary
	if(neg = (n < 0)) n = -n;					// Save sign
	lcd.print(0, r, r + '0');					// Print register #
	lcd.print(12, r, hex[(n >> 12) & 0x0F]);	// Hex 1st digit
	lcd.print(hex[(n >>  8) & 0x0F]);			// 2nd digit
	lcd.print(hex[(n >>  4) & 0x0F]);			// 3rd digit
	lcd.print(hex[(n >>  0) & 0x0F]);			// 4th digit
	div_t d = div(n, 10); lcd.print(72, r, d.rem + '0'); // Decimal 5th digit
	d = div(d.quot, 10); lcd.print(66, r, d.rem + '0');  // 4th digit
	d = div(d.quot, 10); lcd.print(60, r, d.rem + '0');  // 3rd digit
	d = div(d.quot, 10); lcd.print(54, r, d.rem + '0');  // 2nd digit
	lcd.print(48, r, d.quot+ '0');				// 1st digit
	lcd.print(42, r, neg ? '-' : '+');			// Sign
}

int main(void)
{
	WDTCTL = WDTPW | WDTHOLD;					// Disable watchdog reset
	
	BCSCTL1 = CALBC1_1MHZ;						// Run at 1 MHz
	DCOCTL  = CALDCO_1MHZ;						//
	
												// Configure GPIO ports
	P1REN = RXD | SWITCH;
	P1DIR = LED | TXD | LCD_CLK | LCD_BACKLIGHT | LCD_DATA;
	P1OUT = RXD | TXD | SWITCH | LCD_CLK | LCD_BACKLIGHT;
	
	P2DIR = LCD_DC | LCD_CE;
	P2OUT = LCD_DC | LCD_CE;

	iic.init();									// Init IIC
	
	lcd.reset();								// Init LCD
	lcd.init();									//
	lcd.clear();								//

	lcd.printv(0, 0, "MSP430");					// Power up logo
	lcd.printv(78, 0, "INA219");				//
	lcd.bitmap(ti, 18, 0);						//
	__delay_cycles(2000000);					//

	int m = -1;									// No mode active
	
	for(;;) {									// for-ever
		if(P1IN & SWITCH) {						// Check switch
												// Show Volts/Amps/Watts if switch not down
												// Read shunt voltage register
												// This is in units of 1 mA when a 0.01 ohm shunt
												//  resistor is used
			const int sv = ina.read(1);
												// Read bus voltage
												// Mask off status bits and put in units of 1 mV
												/// todo: handle overflow flag
			const int bv = (ina.read(2) & 0xFFF0) >> 1;
			
												// Calculate power in mW
			int p = (long)abs(sv) * bv / 1000;
	
												// Check if mode has changed
			if(m != 0) {						//
				m = 0;							// Now in this mode
				lcd.clear();					// Clear LCD
				lcd.bitmap(AVW, 69, 0);			// Show A/V/W legend
			}		
			print_int(bv, 0);					// Print voltage
			print_int(sv, 2);					// Print amperage
			print_int(p, 4);					// Print power
		} else {
												// Show registers if switch is down
			if(m != 1) {						// Check if mode has changed
				m = 1;							// Now in this mode
				lcd.clear();					// Clear LCD
			}
			unsigned n = 6;						// 6 registers to print
			do {								//
				print_reg(--n);					// Print register, decrement count
			} while(n);							//  ...until register 0 has been printed
		} 
	}	
}
